@@ -26,10 +26,10 @@
 #include <locale.h>
 
 #include <glib/gi18n.h>
+#include <glib/gstdio.h>
 #include <gtk/gtk.h>
 #include <libgnome/libgnome.h>
 
-#define DBUS_API_SUBJECT_TO_CHANGE
 #include <dbus/dbus-glib.h>
 #include <dbus/dbus-glib-lowlevel.h>
 
@@ -38,12 +38,16 @@
 
 #define GSD_DBUS_NAME         "org.gnome.SettingsDaemon"
 
+#define GNOME_SESSION_DBUS_NAME      "org.gnome.SessionManager"
+#define GNOME_SESSION_DBUS_OBJECT    "/org/gnome/SessionManager"
+#define GNOME_SESSION_DBUS_INTERFACE "org.gnome.SessionManager"
+
 static char      *gconf_prefix = NULL;
-static gboolean   no_daemon    = TRUE;
+static gboolean   no_daemon    = FALSE;
 static gboolean   debug        = FALSE;
 
 static GOptionEntry entries[] = {
-        {"debug", 0, 0, G_OPTION_ARG_NONE, &debug, "Enable debugging code", NULL },
+        {"debug", 0, 0, G_OPTION_ARG_NONE, &debug, N_("Enable debugging code"), NULL },
         {"no-daemon", 0, 0, G_OPTION_ARG_NONE, &no_daemon, N_("Don't become a daemon"), NULL },
         {"gconf-prefix", 0, 0, G_OPTION_ARG_STRING, &gconf_prefix, N_("GConf prefix from which to load plugin settings"), NULL},
         {NULL}
@@ -71,10 +75,6 @@ acquire_name_on_proxy (DBusGProxy *bus_proxy)
 
         ret = FALSE;
 
-        if (bus_proxy == NULL) {
-                goto out;
-        }
-
         error = NULL;
         res = dbus_g_proxy_call (bus_proxy,
                                  "RequestName",
@@ -134,9 +134,8 @@ get_session_bus (void)
 }
 
 static gboolean
-bus_register (void)
+bus_register (DBusGConnection *bus)
 {
-        DBusGConnection *bus;
         DBusGProxy      *bus_proxy;
         gboolean         ret;
 
@@ -144,33 +143,69 @@ bus_register (void)
 
         ret = FALSE;
 
-        bus = get_session_bus ();
-        if (bus == NULL) {
-                g_warning ("Could not get a connection to the bus");
-                goto out;
-        }
-
         bus_proxy = get_bus_proxy (bus);
+
         if (bus_proxy == NULL) {
                 g_warning ("Could not construct bus_proxy object");
                 goto out;
         }
 
-        if (! acquire_name_on_proxy (bus_proxy)) {
+        ret = acquire_name_on_proxy (bus_proxy);
+        g_object_unref (bus_proxy);
+
+        if (!ret) {
                 g_warning ("Could not acquire name");
                 goto out;
         }
 
         g_debug ("Successfully connected to D-Bus");
 
-        ret = TRUE;
-
  out:
         gnome_settings_profile_end (NULL);
 
         return ret;
 }
 
+static void
+on_session_over (DBusGProxy *proxy, GnomeSettingsManager *manager)
+{
+        gnome_settings_manager_stop (manager);
+        gtk_main_quit ();
+}
+
+static void
+set_session_over_handler (DBusGConnection *bus, GnomeSettingsManager *manager)
+{
+        DBusGProxy *session_proxy;
+
+        g_assert (bus != NULL);
+
+        gnome_settings_profile_start (NULL);
+
+        session_proxy = 
+                 dbus_g_proxy_new_for_name (bus,
+                                            GNOME_SESSION_DBUS_NAME,
+                                            GNOME_SESSION_DBUS_OBJECT,
+                                            GNOME_SESSION_DBUS_INTERFACE);
+
+	dbus_g_object_register_marshaller (
+		g_cclosure_marshal_VOID__VOID,
+		G_TYPE_NONE,
+		G_TYPE_INVALID);
+
+        dbus_g_proxy_add_signal (session_proxy, 
+                                 "SessionOver",
+                                 G_TYPE_INVALID);
+
+        dbus_g_proxy_connect_signal (session_proxy, 
+                                     "SessionOver",
+                                     G_CALLBACK (on_session_over), 
+                                     manager,
+                                     NULL);
+
+        gnome_settings_profile_end (NULL);
+}
+
 static void
 gsd_log_default_handler (const gchar   *log_domain,
                          GLogLevelFlags log_level,
@@ -194,8 +229,10 @@ main (int argc, char *argv[])
 {
         GnomeSettingsManager *manager;
         GnomeProgram         *program;
+        DBusGConnection      *bus;
         gboolean              res;
         GError               *error;
+        gboolean              create_dirs;
 
         manager = NULL;
         program = NULL;
@@ -214,7 +251,7 @@ main (int argc, char *argv[])
         error = NULL;
         if (! gtk_init_with_args (&argc, &argv, NULL, entries, NULL, &error)) {
                 if (error != NULL) {
-                        g_warning ("%s", error->message);
+                        g_warning (error->message);
                         g_error_free (error);
                 } else {
                         g_warning ("Unable to initialize GTK+");
@@ -229,17 +266,30 @@ main (int argc, char *argv[])
                 g_error ("Could not daemonize: %s", g_strerror (errno));
         }
 
-        if (! bus_register ()) {
+        bus = get_session_bus ();
+        if (bus == NULL) {
+                g_warning ("Could not get a connection to the bus");
                 goto out;
         }
 
+        if (! bus_register (bus)) {
+                goto out;
+        }
+
+        /* If the user does not have a writable HOME directory, then
+           init libgnome with appropriate arguments to run without
+           needing one. */
+        create_dirs = (g_access (g_get_home_dir(), W_OK) != 0);
+
         gnome_settings_profile_start ("gnome_program_init");
         program = gnome_program_init (PACKAGE,
                                       VERSION,
                                       LIBGNOME_MODULE,
                                       argc,
                                       argv,
-                                      GNOME_PARAM_NONE);
+                                      GNOME_PARAM_CREATE_DIRECTORIES,
+                                      create_dirs,
+                                      NULL);
         gnome_settings_profile_end ("gnome_program_init");
 
         gnome_settings_profile_start ("gnome_settings_manager_new");
@@ -250,6 +300,8 @@ main (int argc, char *argv[])
                 goto out;
         }
 
+        set_session_over_handler (bus, manager);
+
         /* If we aren't started by dbus then load the plugins
            automatically.  Otherwise, wait for an Awake etc. */
         if (g_getenv ("DBUS_STARTER_BUS_TYPE") == NULL) {
@@ -268,9 +320,13 @@ main (int argc, char *argv[])
 
         gtk_main ();
 
+ out:
         g_free (gconf_prefix);
 
- out:
+        if (bus != NULL) {
+                dbus_g_connection_unref (bus);
+        }
+
         if (manager != NULL) {
                 g_object_unref (manager);
         }
